import asyncio
import datetime
import json

import freezegun
import pytest

import kopf
from kopf._cogs.configs.progress import ProgressRecord
from kopf._cogs.structs.ephemera import Memo
from kopf._core.actions.application import WAITING_KEEPALIVE_INTERVAL
from kopf._core.actions.execution import TemporaryError
from kopf._core.actions.progression import HandlerState
from kopf._core.engines.indexing import OperatorIndexers
from kopf._core.intents.causes import HANDLER_REASONS, Reason
from kopf._core.reactor.inventory import ResourceMemories
from kopf._core.reactor.processing import process_resource_event


@pytest.mark.parametrize('cause_reason', HANDLER_REASONS)
@pytest.mark.parametrize('now, delayed_iso, delay', [
    ['2020-01-01T00:00:00', '2020-01-01T00:04:56.789000+00:00', 4 * 60 + 56.789],
], ids=['fast'])
async def test_delayed_handlers_progress(
        registry, settings, handlers, resource, cause_mock, cause_reason,
        assert_logs, k8s_mocked, now, delayed_iso, delay):
    handlers.create_mock.side_effect = TemporaryError("oops", delay=delay)
    handlers.update_mock.side_effect = TemporaryError("oops", delay=delay)
    handlers.delete_mock.side_effect = TemporaryError("oops", delay=delay)
    handlers.resume_mock.side_effect = TemporaryError("oops", delay=delay)

    event_type = None if cause_reason == Reason.RESUME else 'irrelevant'
    cause_mock.reason = cause_reason

    with freezegun.freeze_time(now):
        await process_resource_event(
            lifecycle=kopf.lifecycles.all_at_once,
            registry=registry,
            settings=settings,
            resource=resource,
            indexers=OperatorIndexers(),
            memories=ResourceMemories(),
            memobase=Memo(),
            raw_event={'type': event_type, 'object': {}},
            event_queue=asyncio.Queue(),
        )

    assert handlers.create_mock.call_count == (1 if cause_reason == Reason.CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_reason == Reason.UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_reason == Reason.DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_reason == Reason.RESUME else 0)

    assert k8s_mocked.patch.called

    fname = f'{cause_reason}_fn'
    patch = k8s_mocked.patch.call_args_list[0].kwargs['payload']
    progress = json.loads(patch['metadata']['annotations'][f"kopf.zalando.org/{fname}"])
    assert progress['delayed'] == delayed_iso

    assert_logs([
        "Handler .+ is invoked",
        "Handler .+ failed temporarily: oops",
    ])


@pytest.mark.parametrize('cause_reason', HANDLER_REASONS)
@pytest.mark.parametrize('now, delayed_iso, delay', [
    ['2020-01-01T00:00:00', '2020-01-01T00:04:56.789000+00:00', 4 * 60 + 56.789],
    ['2020-01-01T00:00:00', '2099-12-31T23:59:59.000000+00:00', WAITING_KEEPALIVE_INTERVAL],
], ids=['fast', 'slow'])
async def test_delayed_handlers_sleep(
        registry, settings, handlers, resource, cause_mock, cause_reason,
        assert_logs, k8s_mocked, now, delayed_iso, delay, looptime):

    # Any "future" time works and affects nothing as long as it is the same
    basetime = datetime.datetime.now(tz=datetime.timezone.utc)

    # Simulate the original persisted state of the resource.
    # Make sure the finalizer is added since there are mandatory deletion handlers.
    record = ProgressRecord(started='2000-01-01T00:00:00', delayed=delayed_iso)  # a long time ago
    state_dict = HandlerState.from_storage(record, basetime=basetime).as_in_storage()
    event_type = None if cause_reason == Reason.RESUME else 'irrelevant'
    event_body = {
        'metadata': {'finalizers': [settings.persistence.finalizer]},
        'status': {'kopf': {'progress': {
            'create_fn': state_dict,
            'update_fn': state_dict,
            'delete_fn': state_dict,
            'resume_fn': state_dict,
        }}}
    }
    cause_mock.reason = cause_reason

    with freezegun.freeze_time(now):
        await process_resource_event(
            lifecycle=kopf.lifecycles.all_at_once,
            registry=registry,
            settings=settings,
            resource=resource,
            indexers=OperatorIndexers(),
            memories=ResourceMemories(),
            memobase=Memo(),
            raw_event={'type': event_type, 'object': event_body},
            event_queue=asyncio.Queue(),
        )

    assert not handlers.create_mock.called
    assert not handlers.update_mock.called
    assert not handlers.delete_mock.called
    assert not handlers.resume_mock.called

    # The dummy patch is needed to trigger the further changes. The value is irrelevant.
    assert k8s_mocked.patch.called
    patch = k8s_mocked.patch.call_args_list[0].kwargs['payload']
    assert 'kopf.zalando.org/touch-dummy' in patch['metadata']['annotations']

    # The duration of sleep should be as expected.
    assert looptime == delay

    assert_logs([
        r"Sleeping for ([\d\.]+|[\d\.]+ \(capped [\d\.]+\)) seconds",
    ])



import asyncio
from unittest.mock import Mock

import pytest

from kopf._cogs.aiokits.aiotasks import Scheduler


async def f(mock, *args):
    try:
        mock('started')
        for arg in args:
            if isinstance(arg, asyncio.Event):
                arg.set()
            elif isinstance(arg, (int, float)):
                await asyncio.sleep(arg)
            elif callable(arg):
                arg()
    except asyncio.CancelledError:
        mock('cancelled')
    else:
        mock('finished')


async def test_empty_scheduler_lifecycle(looptime):
    scheduler = Scheduler()
    assert scheduler.empty()
    await scheduler.wait()
    assert scheduler.empty()
    await scheduler.close()
    assert scheduler.empty()
    assert looptime == 0


async def test_task_spawning_and_graceful_finishing(looptime):
    mock = Mock()
    flag1 = asyncio.Event()
    flag2 = asyncio.Event()
    scheduler = Scheduler()

    result = await scheduler.spawn(f(mock, flag1, 123, flag2))
    assert result is None

    await flag1.wait()
    assert looptime == 0
    assert mock.call_args_list[0].args[0] == 'started'

    await flag2.wait()
    assert looptime == 123
    assert mock.call_args_list[1].args[0] == 'finished'

    await scheduler.close()


async def test_task_spawning_and_cancellation(looptime):
    mock = Mock()
    flag1 = asyncio.Event()
    flag2 = asyncio.Event()
    scheduler = Scheduler()

    result = await scheduler.spawn(f(mock, flag1, 123, flag2))
    assert result is None

    await flag1.wait()
    assert looptime == 0
    assert mock.call_args_list[0].args[0] == 'started'

    await scheduler.close()
    assert looptime == 0
    assert mock.call_args_list[1].args[0] == 'cancelled'


async def test_no_tasks_are_accepted_after_closing():
    scheduler = Scheduler()
    await scheduler.close()
    assert scheduler._closed
    assert scheduler._spawning_task.done()
    assert scheduler._cleaning_task.done()
    with pytest.raises(RuntimeError, match=r"Cannot add new coroutines"):
        await scheduler.spawn(f(Mock(), 123))


async def test_successes_are_not_reported():
    exception_handler = Mock()
    scheduler = Scheduler(exception_handler=exception_handler)
    await scheduler.spawn(f(Mock()))
    await scheduler.wait()
    await scheduler.close()
    assert exception_handler.call_count == 0


async def test_cancellations_are_not_reported():
    exception_handler = Mock()
    mock = Mock(side_effect=asyncio.CancelledError())
    scheduler = Scheduler(exception_handler=exception_handler)
    await scheduler.spawn(f(mock, 1))
    await scheduler.wait()
    await scheduler.close()
    assert exception_handler.call_count == 0


async def test_exceptions_are_reported():
    exception = ValueError('hello')
    exception_handler = Mock()
    mock = Mock(side_effect=exception)
    scheduler = Scheduler(exception_handler=exception_handler)
    await scheduler.spawn(f(mock))
    await scheduler.wait()
    await scheduler.close()
    assert exception_handler.call_count == 1
    assert exception_handler.call_args.args[0] is exception


async def test_tasks_are_parallel_if_limit_is_not_reached(looptime):
    """
    time:  ////////----------------------0.1s------------------0.2s--///
    task1: ->spawn->start->sleep->finish->|
    task2: ->spawn->start->sleep->finish->|
    """
    task1_started = asyncio.Event()
    task1_finished = asyncio.Event()
    task2_started = asyncio.Event()
    task2_finished = asyncio.Event()
    scheduler = Scheduler(limit=2)

    await scheduler.spawn(f(Mock(), task1_started, 9, task1_finished))
    await scheduler.spawn(f(Mock(), task2_started, 9, task2_finished))
    assert looptime == 0  # i.e. spawning is not not blocking

    await task1_finished.wait()
    assert task2_started.is_set()
    await task2_finished.wait()
    assert looptime == 9

    await scheduler.close()


async def test_tasks_are_pending_if_limit_is_reached(looptime):
    """
    time:  ////////----------------------0.1s------------------0.2s--///
    task1: ->spawn->start->sleep->finish->|
    task2: ->spawn->.....(pending)......->start->sleep->finish->|
    """
    task1_started = asyncio.Event()
    task1_finished = asyncio.Event()
    task2_started = asyncio.Event()
    task2_finished = asyncio.Event()
    scheduler = Scheduler(limit=1)

    await scheduler.spawn(f(Mock(), task1_started, 9, task1_finished))
    await scheduler.spawn(f(Mock(), task2_started, 9, task2_finished))
    assert looptime == 0  # i.e. spawning is not not blocking

    await task1_finished.wait()
    assert not task2_started.is_set()
    await task2_finished.wait()
    assert looptime == 18

    await scheduler.close()



"""
K8s API errors.

The underlying client library (now, ``aiohttp``) can be replaced in the future.
We cannot rely on embedding its exceptions all over the code in the framework.
Hence, we have our own hierarchy of exceptions for K8s API errors.

Low-level errors, such as the network connectivity issues, SSL/HTTPS issues,
etc, are escalated from the client library as is, since they are related not
to the domain of K8s API, but rather to the networking and encryption.

The original errors of the client library are chained as the causes of our own
specialised errors --- for better explainability of errors in the stack traces.

Some selected reasons of K8s API errors are made into their own classes,
so that they can be intercepted and handled in other parts of the framework.
All other reasons are raised as the base error class and are indistinguishable
from each other (except via the exception's fields).

Unlike the underlying client library's errors, the K8s API errors contain more
information about the reasons --- as provided by K8s API in its response bodies,
not guessed only by HTTP statuses alone.

These errors are not exposed to the users, and the users cannot catch them
with ``except:`` clauses. The users can only see these errors in the logs
as the reasons of failures. However, the errors are exposed to other packages.
"""
import json
from collections.abc import Collection
from typing import Literal, TypedDict

import aiohttp

# How many characters of the non-JSON (textual) API errors to log.
# 256 is an arbitrary size based on the gut feeling of what is good for logs & enough for debugging.
# Plus a little hope that no sensitive information usually goes in the leading 256 characters.
# It is not a guarantee, but reduces the probabilities of undesired consequences.
TEXT_ERROR_MAX_SIZE = 256


class RawStatusCause(TypedDict):
    field: str
    reason: str
    message: str


class RawStatusDetails(TypedDict):
    name: str
    uid: str
    retryAfterSeconds: int
    kind: str
    group: str
    causes: Collection[RawStatusCause]


# https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.19/#status-v1-meta
class RawStatus(TypedDict):
    apiVersion: str
    kind: Literal["Status"]
    code: int
    status: Literal["Success", "Failure"]
    reason: str
    message: str
    details: RawStatusDetails


class APIError(Exception):

    def __init__(
            self,
            payload: RawStatus | str | None = None,
            *,
            status: int,
            headers: dict[str, str],
    ) -> None:
        if isinstance(payload, str):
            super().__init__(payload)
        elif payload:
            super().__init__(payload.get('message'), payload)
        else:
            super().__init__()
        self._status = status
        self._headers = headers
        self._payload = payload

    def __repr__(self) -> str:
        subreprs = [repr(arg) for arg in self.args]
        subreprs.append(f'status={self._status!r}')
        return f"{self.__class__.__name__}({', '.join(subreprs)})"

    @property
    def status(self) -> int:
        return self._status

    @property
    def headers(self) -> dict[str, str]:
        return self._headers

    @property
    def code(self) -> int | None:
        return self._payload.get('code') if isinstance(self._payload, dict) else None

    @property
    def message(self) -> str | None:
        return self._payload.get('message') if isinstance(self._payload, dict) else None

    @property
    def details(self) -> RawStatusDetails | None:
        return self._payload.get('details') if isinstance(self._payload, dict) else None


class APIClientError(APIError):  # all 4xx
    pass


class APIServerError(APIError):  # all 5xx
    pass


class APIUnauthorizedError(APIClientError):
    pass


class APIForbiddenError(APIClientError):
    pass


class APINotFoundError(APIClientError):
    pass


class APIConflictError(APIClientError):
    pass


class APIUnprocessableEntityError(APIClientError):
    pass


class APITooManyRequestsError(APIClientError):
    pass


class APISessionClosed(Exception):
    """
    A helper to escalate from inside the requests to cause re-authentication.

    This happens when credentials expire while multiple concurrent requests
    are ongoing (including their retries, mostly their back-off timeouts):
    one random request will raise HTTP 401 and cause the re-authentication,
    while others will retry their requests with the old session (now closed!)
    and get a generic RuntimeError from aiohttp, thus failing their whole task.
    """
    pass


async def check_response(
        response: aiohttp.ClientResponse,
) -> None:
    """
    Check for specialised K8s errors, and raise with extended information.
    """
    if response.status >= 400:

        # Read the response's body before it is closed by raise_for_status().
        payload: RawStatus | str | None
        try:
            payload = await response.json()
        except (json.JSONDecodeError, aiohttp.ContentTypeError, aiohttp.ClientConnectionError):
            payload = await response.text()
            payload = payload.strip() or None

        # Better be safe: who knows which sensitive information can be dumped unless kind==Status.
        if isinstance(payload, dict) and payload.get('kind') != 'Status':
            payload = None

        # Better be safe: if a data blob (not an error) is dumped, protect the logs from overflows.
        if isinstance(payload, str) and len(payload) >= TEXT_ERROR_MAX_SIZE:
            payload = payload[:TEXT_ERROR_MAX_SIZE-3] + '...'

        cls = (
            APIUnauthorizedError if response.status == 401 else
            APIForbiddenError if response.status == 403 else
            APINotFoundError if response.status == 404 else
            APIConflictError if response.status == 409 else
            APIUnprocessableEntityError if response.status == 422 else
            APITooManyRequestsError if response.status == 429 else
            APIClientError if 400 <= response.status < 500 else
            APIServerError if 500 <= response.status < 600 else
            APIError
        )

        # Raise the framework-specific error while keeping the original error in scope.
        # This call also closes the response's body, so it cannot be read afterwards.
        try:
            response.raise_for_status()
        except aiohttp.ClientResponseError as e:
            raise cls(payload, status=response.status, headers=dict(response.headers)) from e
